"""
Defines tools to make ``logging`` comptaible with ``multiprocessing``.

Based largely on: https://gist.github.com/schlamar/7003737

"""

import logging
import os

# Define the formatter for logging messages.
try:
    f = '\n' + '='*os.get_terminal_size().columns + '\n\n'
except OSError as ex:
    # When testing os.get_terminal_size() will fail because stdout is
    # not connceted to a terminal.
    f = '\n' + '='*100 + '\n\n'
formatter = logging.Formatter(fmt=f+('%(asctime)s - %(levelname)s - '
                                     '%(name)s - %(message)s'),
                              datefmt='%H:%M:%S')


# Define logging handlers.
errorhandler = logging.FileHandler('output/scratch/errors.log',
                                   delay=True)
errorhandler.setLevel(logging.ERROR)

streamhandler = logging.StreamHandler()

errorhandler.setFormatter(formatter)
streamhandler.setFormatter(formatter)


class MPLogger(logging.Logger):
    """

    """

    log_queue = None

    def isEnabledFor(self, level):
        return True

    def handle(self, record):
        """

        """

        ei = record.exc_info
        if ei:
            # To get traceback text into record.exc_text
            logging._defaultFormatter.format(record)
            record.exc_info = None  # Not needed any more.
        d = dict(record.__dict__)
        d['msg'] = record.getMessage()
        d['args'] = None
        self.log_queue.put(d)


def logged_call(log_queue, func, *args, **kwargs):
    """

    """

    MPLogger.log_queue = log_queue
    logging.setLoggerClass(MPLogger)
    # Monkey patch root logger and already defined loggers.
    logging.root.__class__ = MPLogger
    for logger in logging.Logger.manager.loggerDict.values():
        if not isinstance(logger, logging.PlaceHolder):
            logger.__class__ = MPLogger
    return func(*args, **kwargs)


def daemon_logger(log_queue):
    """
    Logs messages from subprocesses in the main process.

    Parameters
    ----------
    log_queue : :class:`multiprocessing.Queue`
        A queue into which log records are sent by subprocesses.

    Returns
    -------
    None : :class:`NoneType`

    """

    while True:
        try:
            record_data = log_queue.get()
            if record_data is None:
                break
            record = logging.makeLogRecord(record_data)

            logger = logging.getLogger(record.name)
            if logger.isEnabledFor(record.levelno):
                logger.handle(record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except EOFError:
            break
        except Exception:
            logging.exception('Error in log handler.')



"""
Defines COF topologies.

.. _`cof assembly`:

How the assembly process works for 2D COF topologies.
-----------------------------------------------------

Linker topologies.
..................

Linker topologies are those which use a building block with two
functional groups and a building block with three or more functional
groups. These are implemented as subclasses to
:class:`LinkerCOFLattice`.

As for any assembly process, there are 3 stages: placement of
building blocks, joining of building blocks and deletion of extra
atoms.

A 2D COF topology can be defined by the vertices and edges which
connected them. The building block with 3 or more functional groups
sits on the vertices while the building block with 2 functional groups
sits on the edges. To assemble a COF topology, first the vertices
and edges are defined. They are represented by the classes
:class:`Vertex` and :class:`Edge`. Each of these objects defines their
positions in term of the fractional coordinates along the
``a``, ``b`` and ``c`` vectors of a unit cell. In addtion, each
of these objects defines a set of "atomic positiions".

For edges, there are two atomic positions, 0 and 1. 0 is the position
of the atom which gets connected to an atom on the first vertex
connected by the edge, while 1 is the
position of the atom which gets connected to the second vertex
connected by the edge. Identification of which bonder atom of a
di-functinalized building block sits on which atom position is done
in the following way. Check the distance of each bonder atom to the
first vertex. The closer one is on position 0 the further one is on
position 1.

For vertices there are multiple atomic positions. The number of
atomic positions on a vertex is equal to the number of functional
groups the building block placed on the vertex has. The atomic
positions are labelled from ``0`` to ``n`` where ``n`` is one less than
the number of functional groups in the building block. Going clockwise
around the vertex, starting at the 12 o clock position, the positions
are labelled starting at 0. After placing and aligning a building block
on a vertex, it is necessary to identify which bonder atom sits on
which atomic position. Alignment is done by rotating the builing block
around the z axis so that the distance of one of the bonder atoms
and one of the edges is minimized. Because each edge keeps a record
of which position it connects, the aligned bonder can be assigned to
a position. Then the bonder atoms are assigned to the next positions
in clockwise order.

To join molecules simply go through all the edges in a topology. Each
edge holds the vertices it is connected to and the positions on those
vertices it connects. Each edge and vertex defines
a dictionary which maps the position to the bonder atom which sits on
it. For each edge take the bonder at position ``0`` and the atomic
position on the vertex which is connceted to it. Use the atomic
position on the connected vertex to get the id of the bonder atom which
is to be connected. The same can be done for the bonder at position
``1`` on the edge.

Periodic bonds are not added to the rdkit molelecule, instead they
are registered in :attr:`.Periodic.periodic_bonds`. The method
:meth:`COFLattice.del_atoms` is slightly modified, so that the
coordinates of any deleted atoms are saved. This is necessary for
restoring those atoms when using :meth:`.Periodic.island`.

When creating periodic bonds during :meth:`LinkerCOFLattice.join_mols`
the ids of the bonder atoms involved are saved as the indices within
:attr:`.MacroMolecule.bonder_ids`. This is because deleting atoms
will change the atom ids but not the ordering in
:attr:`~.MacroMolecule.bonder_ids`. However, these are automatically
updated to bonder ids after assembly is completed by
:meth:`.Perdiodic.save_ids`. This is run automatically after assembly
and does not impact implementation but is noted here for completeness.

"""

import rdkit.Chem.AllChem as rdkit
import numpy as np
from scipy.spatial.distance import euclidean
from collections import deque, defaultdict

from .base import Topology
from ...convenience_tools import (PeriodicBond,
                                  add_fragment_props,
                                  normalize_vector)


class Vertex:
    """
    Represents a vertex of a 2D COF topology.

    Attributes
    ----------
    frac_coord : :class:`tuple` of :class:`float`
        The fractional positions along the ``a``, ``b`` and ``c``
        vectors of the vector.

    connected : :class:`list` of :class:`Edge`
        The edges connected to this vertex.

    bonder_map : :class:`dict`
        Maps the id of a bonder position on the vertex to the atom
        id of the bonder sitting on it.

    """

    def __init__(self, frac_coord):
        self.frac_coord = frac_coord
        self.connected = []

    def place_mol(self, cell_params, mol, aligner):
        """
        Places and aligns a molecule on the vertex.

        Parameters
        ----------
        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        mol : :class:`.StructUnit3`
            The building block to be placed.

        aligner : :class:`int`
            The index of a bonder atom in
            :attr:`.StructUnit.bonder_ids` of `mol`. This is the
            bonder atom which gets aligned with an edge.

        Returns
        -------
        :class:`rdkit.Chem.rdchem.Mol`
            The rdkit instance of the placed `mol`.

        """

        coord = self.calc_coord(cell_params)
        original_position = mol.position_matrix()
        mol.set_orientation2([0, 0, 1])

        mol.set_bonder_centroid(coord)
        aligner_edge = next((e for e in self.connected if
                             all(b == 0 for b in e.bond)),
                            self.connected[0])
        vector = (aligner_edge.calc_coord(cell_params) - coord)
        atom = mol.bonder_ids[aligner]
        mol.minimize_theta2(atom, vector, [0, 0, 1])

        rdkit_mol = rdkit.Mol(mol.mol)
        mol.set_position_from_matrix(original_position)
        return rdkit_mol

    def calc_coord(self, cell_params, cell_position=[0, 0, 0]):
        """
        Calculate the position of the vertex within a unit cell.

        Parameters
        ----------
        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        cell_position : :class:`list` of :class:`int`
            Indicates if the vertex is in a periodic cell or not.
            Analogous to the defintion of a periodic bond.

        Returns
        -------
        :class:`numpy.array`
            The position.

        """

        coord = np.zeros((3, ))
        for frac, dim, p in zip(self.frac_coord, cell_params, cell_position):
            coord += (frac+p) * dim

        return coord

    def create_bonder_map(self,
                          macro_mol,
                          cell_params,
                          nbonders,
                          aligned_bonder):
        """
        Creates the attribute :attr:`bonder_map`.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being built.

        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        nbonders : :class:`int`
            The number of bonder atoms the building block placed on the
            vertex has.

        aligned_bonder : :class:`int`
            The index of a bonder atom in
            :attr:`.StructUnit.bonder_ids` of `mol`. This is the
            bonder atom which gets aligned with an edge.

        Returns
        -------
        None : :class:`NoneType`

        """

        center = self.calc_coord(cell_params)

        # Get all the bonder ids.
        bonder_ids = deque(maxlen=nbonders)
        for atom in macro_mol.mol.GetAtoms():
            if atom.HasProp('bonder'):
                bonder_ids.append(atom.GetIdx())

        start = np.array([0, 1])
        angles = []
        for bonder in bonder_ids:
            bonder_coords = macro_mol.atom_coords(bonder) - center
            x, y, _ = normalize_vector(bonder_coords)
            angle = np.arccos(start@np.array([x, y]))
            if x < 0:
                angle = 2*np.pi - angle

            angles.append((angle, bonder))
        angles.sort()

        bonders = [bonder for angle, bonder in angles]
        aligned = bonder_ids[aligned_bonder]
        bonders = (bonders[bonders.index(aligned):] +
                   bonders[:bonders.index(aligned)])

        aligned_pos = self.aligned_position()
        positions = (list(range(aligned_pos, nbonders)) +
                     list(range(0, aligned_pos)))

        self.bonder_map = {}
        for bonder, position in zip(bonders, positions):
            self.bonder_map[position] = bonder

    def aligned_position(self):
        """
        Returns the position of the atom aligned by :meth:`place_mol`.

        Returns
        -------
        :class:`int`
            The id of a position of on the vertex. The bonder atom
            which gets aligned during :meth:`place_mol` sits on this
            position.

        """

        aligner_edge = next((e for e in self.connected if
                             all(b == 0 for b in e.bond)),
                            self.connected[0])
        vindex = 0 if self is aligner_edge.v1 else 1
        return aligner_edge.joint_positions[vindex]


class Edge:
    """
    Represents a bond in a COF unit cell.

    This class is used for positioning and aligning ditopic building
    blocks in 2D COF topologies.

    Attributes
    ----------
    v1 : :class:`Vertex`
        The first vertex connected by the edge.

    v2 : :class:Vertex`
        The second vertex connected by the edge.

    joint_positions : :class:`tuple` of :class:`int`
        Holds 2 numbers. The first is the id of the position on
        :attr:`v1` which gets connected by the edge. The second is the
        id of the position on :attr:`v2` which gets connected by the
        edge.

    bond : :class:`list` of :class:`int`:
        Indicates if the bond is periodic.

    connected : class:`list` of :class:`Vertex`
        The vertices conncetd by the edge.

    frac_coord : :class:`list` of :class`float`
        The position of the edge in terms of the fractional positions
        along the ``a``, ``b`` and ``c`` vectors of a unit cell.

    bonder_map : :class:`dict`
        Maps the id of a position to the id of the bonder atom which
        sits on it. The positions are ``0`` and ``1`` where ``0`` is
        the  position which is closer to :attr:`v1` while ``1``
        is the position closer to :attr:`v2`.

    """

    def __init__(self, v1, v2, joint_positions, bond=[0, 0, 0]):
        self.v1 = v1
        self.v2 = v2
        self.joint_positions = joint_positions
        self.bond = bond
        self.connected = [v1, v2]

        self.frac_coord = []
        for f1, f2, b in zip(v1.frac_coord, v2.frac_coord, bond):
            self.frac_coord.append((f1+f2+b) / 2)

        v1.connected.append(self)
        v2.connected.append(self)

    def place_mol(self, macro_mol, cell_params, mol, alignment):
        """
        Places and aligned a building block along the edge.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being built.

        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        mol : :class:`.StructUnit3`
            The building block to be placed.

        alignment : :class:`int`
            Can be ``1`` or ``-1`` to align `mol` either parallel or
            anti-parallel with the edge.

        Returns
        -------
        :class:`rdkit.Chem.rdchem.Mol`
            The rdkit instance of the placed `mol`.

        """

        coord = self.bonder_centroid(macro_mol, cell_params)
        original_position = mol.position_matrix()

        mol.set_bonder_centroid(coord)
        mol.set_orientation2(self.direction(cell_params)*alignment)

        rdkit_mol = rdkit.Mol(mol.mol)
        mol.set_position_from_matrix(original_position)
        return rdkit_mol

    def direction(self, cell_params):
        """
        The direction vector of the edge.

        Parameters
        ----------
        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        Returns
        -------
        :class:`numpy.array`
            The direction vector.

        """

        v1coord = self.v1.calc_coord(cell_params)
        v2coord = self.v2.calc_coord(cell_params, self.bond)
        return normalize_vector((v1coord-v2coord)/2)

    def calc_coord(self, cell_params):
        """
        Calcualtes the position of the edge in a cell.

        Parameters
        ----------
        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        Returns
        -------
        :class:`numpy.array`
            The position.
        """

        coord = np.zeros((3, ))
        for frac, dim in zip(self.frac_coord, cell_params):
            coord += frac * dim
        return coord

    def bonder_centroid(self, macro_mol, cell_params):
        """
        The centroid of the bonder molecules connected to the edge.

        So the bonders NOT sitting on the edge itself.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being assembled.

        Returns
        -------
        :class:`numpy.array`
            The centroid of the bonder attoms connected to those on the
            edge.

        """

        coord = np.zeros((3, ))
        for i, position in enumerate(self.joint_positions):
            vertex = self.connected[i]
            bonder = vertex.bonder_map[position]
            coord += macro_mol.atom_coords(bonder)

        for d, param in zip(self.bond, cell_params):
            coord += d*param

        return coord / (i+1)

    def create_bonder_map(self, macro_mol, cell_params):
        """
        Creates the attribute :attr:`bonder_map`.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being built.

        cell_params : :class:`list` of :class:`numpy.array`
            The ``a``, ``b`` and ``c`` vectors of the unit cell.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Get all the bonder ids.
        bonder_ids = deque(maxlen=2)
        for atom in macro_mol.mol.GetAtoms():
            if atom.HasProp('bonder'):
                bonder_ids.append(atom.GetIdx())

        v1coord = self.v1.calc_coord(cell_params)
        bonders = sorted(bonder_ids,
                         key=lambda x: euclidean(
                                             v1coord,
                                             macro_mol.atom_coords(x)))
        self.bonder_map = {0: bonders[0],
                           1: bonders[1]}


def is_bonder(macro_mol, atom_id):
    """
    ``True`` if atom has ``'bonder'`` property.

    Parameters
    ----------
    macro_mol : :class:`.MacroMolecule`
        The macromolecule to which the atom belongs.

    Returns
    -------
    :class:`bool`
        ``True`` if atom has ``'bonder'`` property else ``False``.

    """

    return (True if
            macro_mol.mol.GetAtomWithIdx(atom_id).HasProp('bonder')
            else False)


class COFLattice(Topology):
    """
    A base class for periodic topologies.

    This class behaves almost exactly like
    :class:`.Topology` with only minor additions to suit the
    representation of COF periodic lattices. The :meth:`del_atoms`
    method is extended to collect the coordinates of any deleted atoms.
    This is necessary for positioning terminating atoms on islands
    generated from the periodic structure by :meth:`.Periodic.island`.

    """

    def del_atoms(self, macro_mol):
        """
        Deletes the atoms which are lost during assembly.

        Notes
        -----
        The parameter `macro_mol` has two attributes changed.
        :attr:`~.MacroMolecule.mol` has deleter atoms removed, while
        :attr:`~.Periodic.deleters` is updated with the
        coordinates, element type and bond type of every removed atom.

        Parameters
        ----------
        macro_mol : :class:`.Periodic`
            The periodic macromolecule being assembled.

        Returns
        -------
        None : :class:`NoneType`

        """

        # This loop checks if an atom is an bonder and if its is finds
        # a neighboring deleter atom. The deleter atom has its
        # coordinates relative to the bonder found and saved in
        # `terminator_coords`.
        macro_mol.deleters = defaultdict(list)
        for atom in macro_mol.mol.GetAtoms():
            if not atom.HasProp('bonder'):
                continue
            for neighbor in atom.GetNeighbors():
                if not neighbor.HasProp('del'):
                    continue
                bi = macro_mol.bonder_ids.index(atom.GetIdx())
                nid = neighbor.GetIdx()
                tcoords = (macro_mol.atom_coords(nid) -
                           macro_mol.atom_coords(atom.GetIdx()))

                bond = macro_mol.mol.GetBondBetweenAtoms(nid, atom.GetIdx())
                macro_mol.deleters[bi].append([tcoords,
                                               neighbor.GetAtomicNum(),
                                               bond.GetBondType()])

        super().del_atoms(macro_mol)
        macro_mol._ids_updated = False


class LinkerCOFLattice(COFLattice):
    """
    For 2D COF topologies which use a linker.

    These are topologies which consist of 2 building blocks. One
    which has two functional groups and one which has 3 or more
    functional groups.

    This class will have to be extended by sublcasses which define
    the vertices and edges which a particular topology consists of.

    Attributes
    ----------
    vertices : :class:`list` of :class:`Vertex`
        Class attribute added when defining a subclass. Holds the
        vertices which make up the topology.

    edges : :class:`list` of :class:`Edge`
        Class attribute added when defining a subclass. Holds the
        edges which make up the topology.

    ditopic_directions : :class:`list` of :class:`int`
        For each edge in the topology, holds ``1`` or ``-1`` to
        indicate if the building block is placed parallel or
        anti-parallel to the edge.

    multitopic_aligners : :class:`list` of :class`int`
        For each vertex in the topology, holds the index of an
        atom in :attr:`.StructUnit.bonder_ids`. It specicifes which
        bonder atom is used for alignment of the building block on the
        vertex.

    """

    def __init__(self,
                 ditopic_directions=None,
                 multitopic_aligners=None):
        super().__init__()

        if ditopic_directions is None:
            ditopic_directions = [1 for i in range(len(self.edges))]
        if multitopic_aligners is None:
            multitopic_aligners = [0 for i in range(len(self.vertices))]

        self.ditopic_directions = ditopic_directions
        self.multitopic_aligners = multitopic_aligners

    def place_mols(self, macro_mol):
        """
        Places the building blocks on the topology.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being assembled.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Create the rdkit molecule of the assembled molecule.
        macro_mol.mol = rdkit.Mol()

        # Identify which building block is ditopic and which is
        # tri or more topic.
        di = next(bb for bb in macro_mol.building_blocks if
                  len(bb.functional_group_atoms()) == 2)
        multi = next(bb for bb in macro_mol.building_blocks if
                     len(bb.functional_group_atoms()) >= 3)
        nbonders = len(multi.functional_group_atoms())

        # Calculate the size of the unit cell by scaling to the size of
        # building blocks.
        size = di.max_diameter()[0] + multi.max_diameter()[0]
        size *= (len(self.vertices)+1)//1.5
        cell_params = [size*p for p in self.cell_dimensions]
        macro_mol.cell_dimensions = cell_params

        # For each vertex in the topology, place a multitopic building
        # block on it. The Vertex object takes care of alignment.

        for i, v in enumerate(self.vertices):
            aligner = self.multitopic_aligners[i]
            mol = v.place_mol(cell_params, multi, aligner)
            add_fragment_props(mol,
                               macro_mol.building_blocks.index(multi),
                               i)
            macro_mol.mol = rdkit.CombineMols(macro_mol.mol, mol)
            macro_mol.bb_counter.update([multi])

            # Save the ids of the bonder atoms in the assembled molecule.
            # This is used when creating bonds later in the assembly
            # process.
            v.create_bonder_map(macro_mol, cell_params, nbonders, aligner)

        for i, e in enumerate(self.edges):
            mol = e.place_mol(macro_mol,
                              cell_params,
                              di,
                              self.ditopic_directions[i])
            add_fragment_props(mol,
                               macro_mol.building_blocks.index(di),
                               i)
            macro_mol.mol = rdkit.CombineMols(macro_mol.mol, mol)
            macro_mol.bb_counter.update([di])
            bonder_ids = deque(maxlen=2)
            for atom in macro_mol.mol.GetAtoms():
                if atom.HasProp('bonder'):
                    bonder_ids.append(atom.GetIdx())

            e.create_bonder_map(macro_mol, cell_params)

        super(macro_mol.__class__, macro_mol).save_ids()

    def join_mols(self, macro_mol):
        """
        Creates bonds between the building blocks.

        Parameters
        ----------
        macor_mol : :class:`.MacroMolecule`
            The macromolecule being assembled.

        Returns
        -------
        None : :class:`NoneType`

        """

        macro_mol.bonds_made = 0
        emol = rdkit.EditableMol(macro_mol.mol)
        for e in self.edges:
            bonder1 = e.bonder_map[0]
            bonder2 = e.v1.bonder_map[e.joint_positions[0]]
            bond_type = self.determine_bond_type(macro_mol,
                                                 bonder1,
                                                 bonder2)
            emol.AddBond(bonder1, bonder2, bond_type)

            bonder3 = e.bonder_map[1]
            bonder4 = e.v2.bonder_map[e.joint_positions[1]]
            if all(b == 0 for b in e.bond):
                bond_type = self.determine_bond_type(macro_mol,
                                                     bonder3,
                                                     bonder4)
                emol.AddBond(bonder3, bonder4, bond_type)
            else:
                macro_mol.periodic_bonds.append(
                    PeriodicBond(macro_mol.bonder_ids.index(bonder3),
                                 macro_mol.bonder_ids.index(bonder4),
                                 e.bond))

            macro_mol.bonds_made += 2

        macro_mol.mol = emol.GetMol()


class NoLinkerCOFLattice(COFLattice):
    ...


class NoLinkerHoneycomb(NoLinkerCOFLattice):
    """
    Represents a hexagonal lattice with 2 tritopic building blocks.

    """

    cell_dimensions = a, b, c = [np.array([1, 0, 0]),
                                 np.array([0.5, 0.866, 0]),
                                 np.array([0, 0, 5/1.7321])]

    vertices = [(a/3 + b/3 + c/2),
                (2*a/3 + 2*b/3 + c/2)]

    def join_mols(self, macro_mol):
        """
        Joins the building blocks in the unit cell.

        Notes
        -----
        The rdkit instance in the :attr:`~.MacroMolecule.mol` attribute
        has bonds added to it. The
        :attr:`~.MacroMolecule.periodic_bonds` attribute is also filled
        with :class:`PeriodicBond` instances.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The unit cell being constructed.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Get the fragments.
        frag1, frag2 = rdkit.GetMolFrags(macro_mol.mol,
                                         sanitizeFrags=False)
        # Get rid of any non bonder atoms.
        frag1 = [x for x in frag1 if is_bonder(macro_mol, x)]
        frag2 = [x for x in frag2 if is_bonder(macro_mol, x)]

        # The fragment with the larger bonder ids has higher x and y
        # values - due to place_mols() implmentation. It is the "top"
        # fragment.
        top = frag1 if frag1[0] > frag2[0] else frag2
        bottom = frag2 if top is frag1 else frag1

        # In the top fragment find the bonder atom with the
        # largest y value and connect it to the bonder atom in the
        # bottom fragment with the lowest y value. Note that the
        # connection must be registered as periodic, hence the
        # directions are 1/-1.
        top_atom = max(top, key=lambda x: macro_mol.atom_coords(x)[1])
        bottom_atom = min(bottom,
                          key=lambda x: macro_mol.atom_coords(x)[1])
        # The bonder atoms are registered by their index within
        # `bonder_ids`. This is because del_atoms will change the
        # atom ids but not the ordering of this list.
        top_atom = macro_mol.bonder_ids.index(top_atom)
        bottom_atom = macro_mol.bonder_ids.index(bottom_atom)
        macro_mol.periodic_bonds.append(
                    PeriodicBond(top_atom, bottom_atom, [0, 1, 0]))
        # Do the same for the x-axis periodic bonds.
        right_atom = max(top,
                         key=lambda x: macro_mol.atom_coords(x)[0])
        left_atom = min(bottom,
                        key=lambda x: macro_mol.atom_coords(x)[0])
        # The bonder atoms are registered by their index within
        # `bonder_ids`. This is because del_atoms will change the
        # atom ids.
        right_atom = macro_mol.bonder_ids.index(right_atom)
        left_atom = macro_mol.bonder_ids.index(left_atom)
        macro_mol.periodic_bonds.append(
                        PeriodicBond(right_atom, left_atom, [1, 0, 0]))

        # For the bond which gets created directly, find the bonder
        # atom in the bottom fragment closest to the position of the
        # top fragment. Create a bond between it and the bonder atom
        # in the top fragment closest to the position of the bottom
        # fragment.
        bottom_bonder = min(bottom, key=lambda x:
                            euclidean(self.vertices[1],
                                      macro_mol.atom_coords(x)))
        top_bonder = min(top, key=lambda x:
                         euclidean(self.vertices[0],
                                   macro_mol.atom_coords(x)))

        emol = rdkit.EditableMol(macro_mol.mol)
        bond_type = self.determine_bond_type(macro_mol,
                                             top_bonder,
                                             bottom_bonder)
        emol.AddBond(top_bonder, bottom_bonder, bond_type)
        macro_mol.mol = emol.GetMol()

    def place_mols(self, macro_mol):
        """
        Places and aligns building blocks in the unit cell.

        Notes
        -----
        This method modifies `macro_mol`. An rdkit molecule of the
        unit cell with the building blocks not joined up is placed in
        the :attr:`~.MacroMolecule.mol` attribute.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The unit cell being constructed.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Make the rdkit molecule.
        macro_mol.mol = rdkit.Mol()
        # Get the building blocks.
        bb1, bb2 = macro_mol.building_blocks
        cell_size = bb1.max_diameter()[0] + bb2.max_diameter()[0]
        self.cell_dimensions = [cell_size*x for x in
                                self.cell_dimensions]
        self.vertices = [cell_size*x for x in self.vertices]
        # Place and set orientation of the first building block.
        bb1.set_bonder_centroid(self.vertices[0])
        bb1.set_orientation2([0, 0, 1])
        bb1.minimize_theta2(bb1.bonder_ids[0], [0, -1, 0], [0, 0, 1])
        # Add to the macromolecule.
        add_fragment_props(bb1.mol,
                           macro_mol.building_blocks.index(bb1),
                           0)
        macro_mol.mol = rdkit.CombineMols(macro_mol.mol, bb1.mol)
        # Place and set orientation of the second building block.
        bb2.set_bonder_centroid(self.vertices[1])
        bb2.set_orientation2([0, 0, 1])
        bb2.minimize_theta2(bb2.bonder_ids[0], [0, 1, 0], [0, 0, 1])
        # Add to the macromolecule.
        add_fragment_props(bb2.mol,
                           macro_mol.building_blocks.index(bb2),
                           0)
        macro_mol.mol = rdkit.CombineMols(macro_mol.mol, bb2.mol)
        # Add the bonder_ids prematurely for this topology. Needed for
        # making supercells - see join_mols(). Using the ``super``
        # version here because Periodic.save_ids() tries to update the
        # atom ids in periodic bonds, which is needed later in the
        # assembly process. Here only saving of the atom ids
        # is needed, which is done by the ``super`` version.
        super(macro_mol.__class__, macro_mol).save_ids()


class Honeycomb(LinkerCOFLattice):
    cell_dimensions = a, b, c = [np.array([1, 0, 0]),
                                 np.array([0.5, 0.866, 0]),
                                 np.array([0, 0, 5/1.7321])]

    vertices = v1, v2 = [Vertex((1/3, 1/3, 1/2)),
                         Vertex((2/3, 2/3, 1/2))]

    edges = [Edge(v1, v2, (0, 2)),
             Edge(v1, v2, (1, 0), [0, -1, 0]),
             Edge(v1, v2, (2, 1), [-1, 0, 0])]


class Hexagonal(LinkerCOFLattice):
    cell_dimensions = a, b, c = [np.array([1, 0, 0]),
                                 np.array([0.5, 0.866, 0]),
                                 np.array([0, 0, 5/1.7321])]

    vertices = v1, v2, v3, v4 = [Vertex((1/4, 1/4, 1/2)),
                                 Vertex((1/4, 3/4, 1/2)),
                                 Vertex((3/4, 1/4, 1/2)),
                                 Vertex((3/4, 3/4, 1/2))]

    edges = [Edge(v1, v2, (0, 3)),
             Edge(v1, v3, (1, 4)),
             Edge(v2, v3, (2, 5)),
             Edge(v2, v4, (1, 4)),
             Edge(v3, v4, (0, 3)),
             Edge(v1, v3, (4, 1), [-1, 0, 0]),
             Edge(v1, v2, (3, 0), [0, -1, 0]),
             Edge(v1, v4, (2, 5), [0, -1, 0]),
             Edge(v3, v2, (2, 5), [1, -1, 0]),
             Edge(v3, v4, (3, 0), [0, -1, 0]),
             Edge(v2, v4, (4, 1), [-1, 0, 0]),
             Edge(v4, v1, (2, 5), [1, 0, 0])]


class Square(LinkerCOFLattice):
    cell_dimensions = a, b, c = [np.array([1, 0, 0]),
                                 np.array([0, 1, 0]),
                                 np.array([0, 0, 1])]

    vertices = v1, = [Vertex((0.5, 0.5, 0.5))]
    edges = [Edge(v1, v1, (1, 3), [1, 0, 0]),
             Edge(v1, v1, (0, 2), [0, 1, 0])]


class Kagome(LinkerCOFLattice):
    cell_dimensions = a, b, c = [np.array([1, 0, 0]),
                                 np.array([0.5, 0.866, 0]),
                                 np.array([0, 0, 5/1.7321])]

    vertices = v1, v2, v3 = [Vertex((1/4, 3/4, 0.5)),
                             Vertex((3/4, 3/4, 1/2)),
                             Vertex((3/4, 1/4, 1/2))]

    edges = [Edge(v1, v2, (0, 3)),
             Edge(v1, v3, (1, 3)),
             Edge(v2, v3, (2, 0)),
             Edge(v1, v2, (2, 1), [-1, 0, 0]),
             Edge(v1, v3, (3, 1), [-1, 1, 0]),
             Edge(v2, v3, (0, 2), [0, 1, 0])]



from distutils.core import setup

setup(name='stk',
      author='Lukas Turcani',
      author_email='lukasturcani93@gmail.com',
      url='https://www.github.com/lukasturcani/stk',
      version='2018.04.11.2',
      packages=['stk',
                'stk.convenience_tools',
                'stk.molecular',
                'stk.molecular.topologies',
                'stk.molecular.topologies.cage',
                'stk.optimization'],
      install_requires=['networkx',
                        'scipy',
                        'matplotlib',
                        'sklearn',
                        'psutil',
                        'pywindowx'],
      python_requires='>=3.6')



"""
Defines cage topologies of building blocks with 3 functional groups.

"""

import numpy as np
import rdkit.Chem.AllChem as rdkit

from .base import _NoLinkerCageTopology,  Vertex


class OnePlusOne(_NoLinkerCageTopology):
    """
    A sandwich cage topology from tri-functionalised building blocks.

    """

    x = 1
    positions_A = [Vertex(x, 0, 0),
                   Vertex(-x, 0, 0)]
    a, b = positions_A
    connections = [(a, b)]

    a.edge_plane_normal = lambda scale: scale*np.array([1, 0, 0])
    b.edge_plane_normal = lambda scale: scale*np.array([-1, 0, 0])

    a.edge_centroid = lambda scale: scale*np.array([0, 0, 0])
    b.edge_centroid = lambda scale: scale*np.array([0, 0, 0])

    n_windows = 3
    n_window_types = 1

    def join_mols(self, macro_mol):

        editable_mol = rdkit.EditableMol(macro_mol.mol)

        for position in self.positions_A:
            other_position = next(x for x in self.positions_A if
                                  x is not position)

            position.atom_position_pairs = [(atom, other_position) for
                                            atom in position.bonder_ids]

            for atom_id, vertex in position.atom_position_pairs:
                # Get all the distances between the atom and the bonder
                # atoms on the vertex. Store this information on the
                # vertex.
                for atom2_id in vertex.bonder_ids:
                    distance = macro_mol.atom_distance(atom_id,
                                                       atom2_id)
                    position.distances.append((distance,
                                              atom_id, atom2_id))

        paired = set()
        for position in self.positions_A:
            for _, atom1_id, atom2_id in sorted(position.distances):
                if atom1_id in paired or atom2_id in paired:
                    continue

                bond_type = self.determine_bond_type(macro_mol,
                                                     atom1_id,
                                                     atom2_id)
                # Add the bond.
                editable_mol.AddBond(atom1_id, atom2_id, bond_type)
                macro_mol.bonds_made += 1
                paired.add(atom1_id)
                paired.add(atom2_id)

        macro_mol.mol = editable_mol.GetMol()


class TwoPlusTwo(_NoLinkerCageTopology):
    """
    Tetrahedral cage topology from tri-functionalised building blocks.

    """

    x = 1
    positions_A = [Vertex(x, 0, -x/np.sqrt(2)),
                   Vertex(-x, 0, -x/np.sqrt(2)),
                   Vertex(0, x, x/np.sqrt(2)),
                   Vertex(0, -x, x/np.sqrt(2))]

    a, b, c, d = positions_A

    for x in positions_A:
        old_normal = x.edge_plane_normal
        x.edge_plane_normal = lambda scale, a=old_normal: -1*a(scale)

    connections = [(a, b), (a, c), (a, d),
                   (b, c), (b, d),
                   (c, d)]

    n_windows = 4
    n_window_types = 1


class FourPlusFour(_NoLinkerCageTopology):
    """
    A square cage topology from tri-functionalised building blocks.

    """

    x = 1
    positions_A = [Vertex(-x, x, -x),
                   Vertex(-x, -x, -x),
                   Vertex(x, x, -x),
                   Vertex(x, -x, -x),

                   Vertex(-x, x, x),
                   Vertex(-x, -x, x),
                   Vertex(x, x, x),
                   Vertex(x, -x, x)]

    a, b, c, d, e, f, g, h = positions_A

    connections = [(a, b), (a, c), (a, e), (b, d), (b, f), (c, g),
                   (c, d), (d, h), (e, g), (e, f), (f, h), (g, h)]

    n_windows = 6
    n_window_types = 1



"""
This module defines general-purpose objects, functions and classes.

Functions, classes, etc. defined here should not depend on any other
part of MMEA. They must be completely self-sufficient.

"""

import rdkit.Chem.AllChem as rdkit
from rdkit.Geometry import Point3D
import numpy as np
import time
from contextlib import contextmanager
import os
import subprocess as sp
import gzip
import re
from collections import deque
import tarfile

# Holds the elements Van der Waals radii in Angstroms.
atom_vdw_radii = {
              'Al': 2, 'Sb': 2, 'Ar': 1.88, 'As': 1.85, 'Ba': 2,
              'Be': 2, 'Bi': 2, 'B': 2, 'Br': 1.85, 'Cd': 1.58,
              'Cs': 2, 'Ca': 2, 'C': 1.7, 'Ce': 2, 'Cl': 1.75,
              'Cr': 2, 'Co': 2, 'Cu': 1.4, 'Dy': 2, 'Er': 2,
              'Eu': 2, 'F':  1.47, 'Gd': 2, 'Ga': 1.87, 'Ge': 2,
              'Au': 1.66, 'Hf': 2, 'He': 1.4, 'Ho': 2, 'H': 1.09,
              'In': 1.93, 'I': 1.98, 'Ir': 2, 'Fe': 2, 'Kr': 2.02,
              'La': 2, 'Pb': 2.02, 'Li': 1.82, 'Lu': 2, 'Mg': 1.73,
              'Mn': 2, 'Hg': 1.55, 'Mo': 2, 'Nd': 2, 'Ne': 1.54,
              'Ni': 1.63, 'Nb': 2, 'N':  1.55, 'Os': 2, 'O':  1.52,
              'Pd': 1.63, 'P': 1.8, 'Pt': 1.72, 'K': 2.75, 'Pr': 2,
              'Pa': 2, 'Re': 2, 'Rh': 2, 'Rb': 2, 'Ru': 2, 'Sm': 2,
              'Sc': 2, 'Se': 1.9, 'Si': 2.1, 'Ag': 1.72, 'Na': 2.27,
              'Sr': 2, 'S': 1.8, 'Ta': 2, 'Te': 2.06, 'Tb': 2,
              'Tl': 1.96, 'Th': 2, 'Tm': 2, 'Sn': 2.17, 'Ti': 2,
              'W': 2, 'U':  1.86, 'V':  2, 'Xe': 2.16, 'Yb': 2,
              'Y': 2, 'Zn': 1.29, 'Zr': 2, 'X':  1.0, 'D':  1.0
                 }

# This dictionary gives easy access to the rdkit bond types.
bond_dict = {'1': rdkit.rdchem.BondType.SINGLE,
             'am': rdkit.rdchem.BondType.SINGLE,
             '2': rdkit.rdchem.BondType.DOUBLE,
             '3': rdkit.rdchem.BondType.TRIPLE,
             'ar': rdkit.rdchem.BondType.AROMATIC}

# A dictionary which matches atomic number to elemental symbols.
periodic_table = {
              1: 'H', 2: 'He', 3: 'Li', 4: 'Be', 5: 'B', 6: 'C',
              7: 'N', 8: 'O',  9: 'F', 10: 'Ne', 11: 'Na', 12: 'Mg',
              13: 'Al', 14: 'Si', 15: 'P', 16: 'S', 17: 'Cl',
              18: 'Ar', 19: 'K', 20: 'Ca', 21: 'Sc', 22: 'Ti',
              23: 'V', 24: 'Cr', 25: 'Mn', 26: 'Fe', 27: 'Co',
              28: 'Ni', 29: 'Cu', 30: 'Zn', 31: 'Ga', 32: 'Ge',
              33: 'As', 34: 'Se', 35: 'Br', 36: 'Kr', 37: 'Rb',
              38: 'Sr', 39: 'Y', 40: 'Zr', 41: 'Nb', 42: 'Mo',
              43: 'Tc', 44: 'Ru', 45: 'Rh', 46: 'Pd', 47: 'Ag',
              48: 'Cd', 49: 'In', 50: 'Sn', 51: 'Sb', 52: 'Te',
              53: 'I', 54: 'Xe', 55: 'Cs', 56: 'Ba', 57: 'La',
              58: 'Ce', 59: 'Pr', 60: 'Nd', 61: 'Pm', 62: 'Sm',
              63: 'Eu', 64: 'Gd', 65: 'Tb', 66: 'Dy', 67: 'Ho',
              68: 'Er', 69: 'Tm', 70: 'Yb', 71: 'Lu', 72: 'Hf',
              73: 'Ta', 74: 'W', 75: 'Re', 76: 'Os', 77: 'Ir',
              78: 'Pt', 79: 'Au', 80: 'Hg', 81: 'Tl', 82: 'Pb',
              83: 'Bi', 84: 'Po', 85: 'At', 86: 'Rn', 87: 'Fr',
              88: 'Ra', 89: 'Ac', 90: 'Th', 91: 'Pa', 92: 'U',
              93: 'Np', 94: 'Pu', 95: 'Am', 96: 'Cm', 97: 'Bk',
              98: 'Cf', 99: 'Es', 100: 'Fm', 101: 'Md', 102: 'No',
              103: 'Lr', 104: 'Rf', 105: 'Db', 106: 'Sg', 107: 'Bh',
              108: 'Hs', 109: 'Mt', 110: 'Ds', 111: 'Rg', 112: 'Cn',
              113: 'Uut', 114: 'Fl', 115: 'Uup', 116: 'Lv',
              117: 'Uus', 118: 'Uuo'}


class Cell:
    """
    Represents an individual cell in a supercell.

    Parameters
    ----------
    id : :class:`list` of :class:`int`
        A 3 member :class:`list` holding the x, y and z index
        of the cell within the supercell.

    bonders : :class:`dict`
        Maps the bonder atoms in the original unit cell to the
        equivalent bonder atoms in the cell.

    Attributes
    ----------
    id : :class:`numpy.ndarray` of :class:`int`
        A 3 member array holding the x, y and z index
        of the cell within the supercell.

    bonders : :class:`dict`
        Maps the bonder atoms in the original unit cell to the
        equivalent bonder atoms in the cell.

    """

    def __init__(self, id_, bonders):
        self.id = np.array(id_)
        self.bonders = bonders


class ChargedMolError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class MolFileError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class PopulationSizeError(Exception):
    def __init__(self, msg):
        self.msg = msg


class FunctionData:
    """
    Stores information about functions and their parameters.

    Attributes
    ----------
    name : str
        The name of a function or method.

    params : dict
        The parameters of the function or method who's name is held by
        `name`.

    """

    __slots__ = ['name', 'params']

    def __init__(self, name, **kwargs):
        self.name = name
        self.params = kwargs

    def __hash__(self):
        return 1

    def __eq__(self, other):

        same_length = len(self) == len(other)
        same_items = all(x in other.params.items() for x in
                         self.params.items())
        same_name = self.name == other.name

        return same_length and same_items and same_name

    def __len__(self):
        return len(self.params.items())

    def __str__(self):
        s = ", ".join("{}={!r}".format(key, value) for key, value in
                      self.params.items())
        return "FunctionData({!r}, ".format(self.name) + s + ")"

    def __repr__(self):
        return str(self)


class LazyAttr:
    """
    A descriptor for creating lazy attributes.

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        val = self.func(obj)
        setattr(obj, self.func.__name__, val)
        return val


class MAEExtractor:
    """
    Extracts the lowest energy conformer from a .maegz file.

    Macromodel conformer searches produce -out.maegz files containing
    all of the conformers found during the search and their energies
    and other data.

    Initializing this class with a MacroMolecule finds that
    MacroMolecules -out.maegz file and converts it to a .mae file. It
    then creates and additional .mae file holding only the lowest
    energy conformer found.

    Attributes
    ----------
    maegz_path : :class:`str`
        The path to the ``-out.maegz`` file generated by the macromodel
        conformer search.

    mae_path : :class:`str`
        The path to the ``.mae`` file holding the conformers generated by
        the macromodel conformer search.

    content : :class:`str`
        The content of the ``.mae`` file hodling all the conformers from
        the macromodel conformer search. This holds other data such as
        their energies too.

    energies : :class:`list`
        The :class:`list` has the form

        .. code-block:: python

            energies = [(0, 231.0), (1, 144.4), ...]

        Each :class:`tuple` holds the id and energy of every conformer in the
        ``.mae`` file, respectively.

    min_energy : :class:`float`
        The minimum energy found in the ``.mae`` file.

    path : :class:`str`
        The full path of the ``.mae`` file holding the extracted lowest
        energy conformer.

    """

    def __init__(self, file, n=1):

        name, ext = os.path.splitext(file)
        self.maegz_path = name + '-out.maegz'
        self.maegz_to_mae()
        self.extract_conformers(n)

    def extract_conformers(self, n):
        """
        Creates a series of ``.mae`` files holding the lowest energy conformers.

        """

        for i in range(n):
            # Get the id of the lowest energy conformer.
            num = self.lowest_energy_conformers(n)[i][1]
            # Get the structure block corresponding to the lowest energy
            # conformer.
            content = self.content.split("f_m_ct")
            new_mae = "f_m_ct".join([content[0], content[num]])

            # Write the structure block in its own .mae file, named after
            # conformer extracted.
            if n == 1:
                # Write the structure block in its own .mae file, named after
                # conformer extracted.
                new_name = self.mae_path.replace(
                                            '.mae',
                                            'EXTRACTED_{}.mae'.format(num))
            else:
                new_name = self.mae_path.replace(
                              '.mae', 'EXTRACTED_{}_conf_{}.mae'.format(num, i))

            with open(new_name, 'w') as mae_file:
                mae_file.write(new_mae)

            if i == 0:
                # Save the path of the newly created file.
                self.path = new_name

    def extract_energy(self, block):
        """
        Extracts the energy value from a .mae energy data block.

        """

        block = block.split(":::")
        for name, value in zip(block[0].split('\n'),
                               block[1].split('\n')):
            if 'r_mmod_Potential_Energy' in name:
                return float(value)

    def lowest_energy_conformers(self, n):
        """
        Returns the id and energy of the lowest energy conformers.

        Parameters
        ----------
        n : :class:`int`
            The number of lowest energy conformers to return.

        Returns
        -------
        :class:`list`
            A :class:`list` of the form

            .. code-block:: python

                returned = [(23, 123.3), (1, 143.89), (12, 150.6), ...]

            Where each :class:`tuple` holds the id and energy of the `n`
            lowest energy conformers, respectively

        """

        # Open the .mae file holding all the conformers and load its
        # content.
        with open(self.mae_path, 'r') as mae_file:
            self.content = mae_file.read()
            # Split the content across curly braces. This divides the
            # various sections of the .mae file.
            content_split = re.split(r"[{}]", self.content)

        # Go through all the datablocks in the the .mae file. For each
        # energy block extract the energy and store it in the
        # `energies` list. Store the `index`  (conformer id) along with
        # each extracted energy.
        self.energies = []
        prev_block = deque([""], maxlen=1)
        index = 1
        for block in content_split:
            if ("f_m_ct" in prev_block[0] and
               "r_mmod_Potential_Energy" in block):
                energy = self.extract_energy(block)
                self.energies.append((energy, index))
                index += 1

            prev_block.append(block)

        # Selecting the lowest energy n conformers
        confs = sorted(self.energies)[:n]
        # Define the energy of the lowest energy conformer
        self.min_energy = confs[0][0]
        # Return a list with id and energy of the lowest energy conformers.
        return confs

    def maegz_to_mae(self):
        """
        Converts the .maegz file to a .mae file.

        """

        self.mae_path = self.maegz_path.replace('.maegz', '.mae')
        with gzip.open(self.maegz_path, 'r') as maegz_file:
            with open(self.mae_path, 'wb') as mae_file:
                mae_file.write(maegz_file.read())


class PeriodicBond:
    """
    Represents a periodic bond.

    Parameters
    ----------
    atom1 : :class:`int`
        The id of first atom in the bond.

    atom2 : :class:`int`
        The id of the second atom in the bond.

    direction : :class:`list` of :class:`int`
        A 3 member list describing the axes along which the bond is
        periodic, when going from `atom1` to `atom2`. For example,
        ``[1, 0, 0]`` means that the bond is periodic along the x axis
        in the positive direction.


    Attributes
    ----------
    atom1 : :class:`int`
        The id of first atom in the bond.

    atom2 : :class:`int`
        The id of the second atom in the bond.

    direction : :class:`numpy.ndarray` of :class:`int`
        A 3 member list describing the axes along which the bond is
        periodic, when going from `atom1` to `atom2`. For example,
        ``[1, 0, 0]`` means that the bond is periodic along the x axis
        in the positive direction.

    """

    def __init__(self, atom1, atom2, direction):
        self.atom1 = atom1
        self.atom2 = atom2
        self.direction = np.array(direction)

    def __str__(self):
        return ("PeriodicBond({0.atom1}, "
                "{0.atom2}, {0.direction})").format(self)


class StopLogging:
    ...


def add_fragment_props(mol, bb_index, mol_index):
    """
    Adds properties to `mol` of `bb_index` and `mol_index`.

    Properties called 'bb_index' and 'mol_index' are added to every
    atom in `mol`.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        A molecule which needs to have its atoms tagged.

    bb_index : int
        The index of `mol` within `building_blocks` of some
        macromolecule.

    mol_index : int
        If `mol` is the 5th molecule of building block with `bb_index`
        to be added to a macromolecule, `mol_index` is 4.

    Returns
    -------
    None : NoneType

    """

    for atom in mol.GetAtoms():
        atom.SetIntProp('bb_index', bb_index)
        atom.SetIntProp('mol_index', mol_index)


def archive_output():
    """
    Places the ``output`` folder into ``old_output``.

    This function assumes that the ``output`` folder is in the current
    working directory.

    Returns
    -------
    None : NoneType

    """

    if 'output' not in os.listdir():
        return

    # Make the ``old_output`` folder if it does not exist already.
    if 'old_output' not in os.listdir():
        os.mkdir('old_output')

    # Find out with what number the ``output`` folder should be
    # labelled within ``old_output``.
    num = len(os.listdir('old_output'))
    new_dir = os.path.join('old_output', str(num))
    os.rename('output', new_dir)


def centroid(*coords):
    """
    Calculates the centroid of a group of coordinates.

    Parameters
    ----------
    *coords : numpy.array
        Any number of numpy arrays holding x, y and z positions.

    Returns
    -------
    numpy.array
        The centroid of the coordinates `coords`.

    """

    total = 0
    for coord in coords:
        total = np.add(total, coord)
    return np.divide(total, len(coords))


def dedupe(iterable, seen=None, key=None):
    """
    Yields items from `iterable` barring duplicates.

    If `seen` is provided it contains elements which are not to be
    yielded at all.

    Parameters
    ----------
    iterable : iterable
        An iterable of elements which are to be yielded, only once.

    seen : set (default = None)
        Holds items which are not to be yielded.

    key : callable
        A function which gets applied to every member of `iterable`.
        The return of this function is checked for duplication rather
        than the member itself.

    Yields
    ------
    object
        Element in `iterable` which is not in `seen` and has not been
        yielded before.

    """

    if seen is None:
        seen = set()
    for x in iterable:
        val = key(x) if key is not None else x
        if val not in seen:
            seen.add(val)
            yield x


def flatten(iterable, excluded_types={str}):
    """
    Transforms an nested iterable into a flat one.

    For example

        [[1,2,3], [[4], [5],[[6]], 7]

    becomes

        [1,2,3,4,5,6,7]

    If a type is found in `excluded_types` it will not be yielded from.
    For example if `str` is in `excluded_types`

        a = ["abcd", ["efgh"]]

    "abcd" and "efgh" are yielded if `a` is passed to `iterable`. If
    `str` was not in `excluded_types` then "a", "b", "c", "d", "e",
    "f", "g" and "h" would all be yielded individually.

    Parameters
    ----------
    iterable : iterable
        The iterable which is to be flattened

    excluded_types : set
        Holds container types which are not be flattened.

    Yields
    ------
    object
        A nested element of `iterable`.

    """

    for x in iterable:
        if hasattr(x, '__iter__') and type(x) not in excluded_types:
            yield from flatten(x)
        else:
            yield x


def kabsch(coords1, coords2):
    """
    Return a rotation matrix to minimize dstance between 2 coord sets.

    This is essentially an implementation of the Kabsch algorithm.
    Given two sets of coordinates, `coords1` and `coords2`, this
    function returns a rotation matrix. When the rotation matrix is
    applied to `coords1` the resulting coordinates have their rms
    distance to `coords2` minimized.

    Parameters
    ----------
    coords1 : numpy.array
        This array represents a matrix hodling coordinates which need
        to be rotated to minimize their rms distance to coordinates in
        `coords2`. The matrix is n x 3. Each row of the matrix holds
        the x, y and z coordinates of one point, respectively. Here
        ``n`` is the number of points.

    coords2 : numpy.array
        This array represents a matrix which holds the coordinates of
        points the distance to which should be minimized. The matrix is
        n x 3. Each row of the matrix holds the x, y and z coordinates
        of one point, respectively. Here ``n`` is the number of points.

    Returns
    -------
    numpy.array
        A rotation matrix. This will be a 3 x 3 matrix.

    References
    ----------
    http://nghiaho.com/?page_id=671
    https://en.wikipedia.org/wiki/Kabsch_algorithm

    """

    h = np.dot(coords1, coords2.T)
    u, s, v = np.linalg.svd(h)

    if int(np.linalg.det(v)) < 0:
        v[:, 2] = -v[:, 2]

    return np.dot(v, u)


def kill_macromodel():
    """
    Kills any applications left open as a result running MacroModel.

    Applications that are typically left open are
    ``jserver-watcher.exe`` and ``jservergo.exe``.

    Returns
    -------
    None : NoneType

    """

    if os.name == 'nt':
        # In Windows, use the ``Taskkill`` command to force a close on
        # the applications.
        sp.run(["Taskkill", "/IM", "jserver-watcher.exe", "/F"],
               stdout=sp.PIPE, stderr=sp.PIPE)
        sp.run(["Taskkill", "/IM", "jservergo.exe", "/F"],
               stdout=sp.PIPE, stderr=sp.PIPE)

    if os.name == 'posix':
        sp.run(["pkill", "jservergo"],
               stdout=sp.PIPE, stderr=sp.PIPE)
        sp.run(["pkill", "jserver-watcher"],
               stdout=sp.PIPE, stderr=sp.PIPE)


def matrix_centroid(matrix):
    """
    Returns the centroid of the coordinates held in `matrix`.

    Parameters
    ----------
    matrix : np.matrix
        A n x 3 matrix. Each row holds the x, y and z coordinate of
        some point, respectively.

    Returns
    -------
    numpy.array
        A numpy array which holds the x, y and z coordinates of the
        centroid of the coordinates in `matrix`.

    """

    return np.array(np.sum(matrix, axis=0) / len(matrix))[0]


def mol_from_mae_file(mae_path):
    """
    Creates a rdkit molecule from a ``.mae`` file.

    Parameters
    ----------
    mol2_file : str
        The full path of the ``.mae`` file from which an rdkit molecule
        should be instantiated.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        An rdkit instance of the molecule held in `mae_file`.

    """

    mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mae_path, 'r') as mae:
        content = re.split(r'[{}]', mae.read())

    prev_block = deque([''], maxlen=1)
    for block in content:
        if 'm_atom[' in prev_block[0]:
            atom_block = block
        if 'm_bond[' in prev_block[0]:
            bond_block = block
        prev_block.append(block)

    labels, data_block, *_ = atom_block.split(':::')
    labels = [l for l in labels.split('\n') if
              not l.isspace() and l != '']

    data_block = [a.split() for a in data_block.split('\n') if
                  not a.isspace() and a != '']

    for line in data_block:
        line = [word for word in line if word != '"']
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of columns'
                                ' in .mae file.'))

        for label, data in zip(labels, line):
            if 'x_coord' in label:
                x = float(data)
            if 'y_coord' in label:
                y = float(data)
            if 'z_coord' in label:
                z = float(data)
            if 'atomic_number' in label:
                atom_num = int(data)

        atom_sym = periodic_table[atom_num]
        atom_coord = Point3D(x, y, z)
        atom_id = mol.AddAtom(rdkit.Atom(atom_sym))
        conf.SetAtomPosition(atom_id, atom_coord)

    labels, data_block, *_ = bond_block.split(':::')
    labels = [l for l in
              labels.split('\n') if not l.isspace() and l != '']
    data_block = [a.split() for a in
                  data_block.split('\n') if not a.isspace() and a != '']

    for line in data_block:
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of '
                                'columns in .mae file.'))

        for label, data in zip(labels, line):
            if 'from' in label:
                atom1 = int(data) - 1
            if 'to' in label:
                atom2 = int(data) - 1
            if 'order' in label:
                bond_order = str(int(data))
        mol.AddBond(atom1, atom2, bond_dict[bond_order])

    mol = mol.GetMol()
    mol.AddConformer(conf)
    return mol


def mol_from_mol_file(mol_file):
    """
    Creates a rdkit molecule from a ``.mol`` (V3000) file.

    Parameters
    ----------
    mol_file : str
        The full of the .mol file from which an rdkit molecule should
        be instantiated.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        An rdkit instance of the molecule held in `mol2_file`.

    Raises
    ------
    ChargedMolError
        If an atom row has more than 8 coloumns it is usually because
        there is a 9th coloumn indicating atomic charge. Such molecules
        are not currently supported, so an error is raised.
    MolFileError
        If the file is not a V3000 .mol file.

    """

    e_mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mol_file, 'r') as f:
        take_atom = False
        take_bond = False
        v3000 = False

        for line in f:
            if 'V3000' in line:
                v3000 = True

            if 'M  V30 BEGIN ATOM' in line:
                take_atom = True
                continue

            if 'M  V30 END ATOM' in line:
                take_atom = False
                continue

            if 'M  V30 BEGIN BOND' in line:
                take_bond = True
                continue

            if 'M  V30 END BOND' in line:
                take_bond = False
                continue

            if take_atom:
                words = line.split()
                if len(words) > 8:
                    raise ChargedMolError(mol_file,
                                          ('Atom row has more'
                                           ' than 8 coloumns. Likely '
                                           'due to a charged atom.'))
                _, _, _, atom_sym, *coords, _ = words
                coords = [float(x) for x in coords]
                atom_coord = Point3D(*coords)
                atom_id = e_mol.AddAtom(rdkit.Atom(atom_sym))
                conf.SetAtomPosition(atom_id, atom_coord)
                continue

            if take_bond:
                *_, bond_id,  bond_order, atom1, atom2 = line.split()
                e_mol.AddBond(int(atom1)-1, int(atom2)-1,
                              bond_dict[bond_order])
                continue
    if not v3000:
        raise MolFileError(mol_file, 'Not a V3000 .mol file.')

    mol = e_mol.GetMol()
    mol.AddConformer(conf)
    return mol


def normalize_vector(vector):
    """
    Normalizes the given vector.

    A new vector is returned, the original vector is not modified.

    Parameters
    ----------
    vector : np.array
        The vector to be normalized.

    Returns
    -------
    np.array
        The normalized vector.

    """

    v = np.divide(vector, np.linalg.norm(vector))
    return np.round(v, decimals=4)


def rotation_matrix(vector1, vector2):
    """
    Returns a rotation matrix which transforms `vector1` to `vector2`.

    Multiplying `vector1` by the rotation matrix returned by this
    function yields `vector2`.

    Parameters
    ----------
    vector1 : numpy.array
        The vector which needs to be transformed to `vector2`.

    vector2 : numpy.array
        The vector onto which `vector1` needs to be transformed.

    Returns
    -------
    numpy.ndarray
        A rotation matrix which transforms `vector1` to `vector2`.

    References
    ----------
    http://tinyurl.com/kybj9ox
    http://tinyurl.com/gn6e8mz

    """

    # Make sure both inputs are unit vectors.
    vector1 = normalize_vector(vector1)
    vector2 = normalize_vector(vector2)

    # Hande the case where the input and output vectors are equal.
    if np.array_equal(vector1, vector2):
        return np.identity(3)

    # Handle the case where the rotation is 180 degrees.
    if np.array_equal(vector1, np.multiply(vector2, -1)):
        # Get a vector orthogonal to `vector1` by finding the smallest
        # component of `vector1` and making that a vector.
        ortho = [0, 0, 0]
        ortho[list(vector1).index(min(abs(vector1)))] = 1
        axis = np.cross(vector1, ortho)
        return rotation_matrix_arbitrary_axis(np.pi, axis)

    v = np.cross(vector1, vector2)

    vx = np.array([[0, -v[2], v[1]],
                   [v[2], 0, -v[0]],
                   [-v[1], v[0], 0]])

    s = np.linalg.norm(v)
    c = np.dot(vector1, vector2)
    I = np.identity(3)

    mult_factor = (1-c)/np.square(s)

    return I + vx + np.multiply(np.dot(vx, vx), mult_factor)


def rotation_matrix_arbitrary_axis(angle, axis):
    """

    Returns a rotation matrix of `angle` radians about `axis`.

    Parameters
    ----------
    angle : int or float
        The size of the rotation in radians.

    axis : numpy.array
        A 3 element aray which represents a vector. The vector is the
        axis about which the rotation is carried out.

    Returns
    -------
    numpy.array
        A 3x3 array representing a rotation matrix.

    """

    axis = normalize_vector(axis)

    a = np.cos(angle/2)
    b, c, d = axis * np.sin(angle/2)

    e11 = np.square(a) + np.square(b) - np.square(c) - np.square(d)
    e12 = 2*(b*c - a*d)
    e13 = 2*(b*d + a*c)

    e21 = 2*(b*c + a*d)
    e22 = np.square(a) + np.square(c) - np.square(b) - np.square(d)
    e23 = 2*(c*d - a*b)

    e31 = 2*(b*d - a*c)
    e32 = 2*(c*d + a*b)
    e33 = np.square(a) + np.square(d) - np.square(b) - np.square(c)

    return np.array([[e11, e12, e13],
                     [e21, e22, e23],
                     [e31, e32, e33]])


def tar_output():
    """
    Places all the content in the `output` folder into a .tgz file.

    Returns
    -------
    None : NoneType

    """

    tname = os.path.join('output', 'output.tgz')
    with tarfile.open(tname, 'w:gz') as tar:
        tar.add('output')


@contextmanager
def time_it():
    """
    Times the code executed within the indent.

    This is a context manager so it should be used as:

        with time_it():
            something1()
            something2()
            something3()

    After all 3 functions are finished and the nested block is exited
    the time taken to process the entire block is printed.

    """

    start = time.time()
    yield
    time_taken = time.time() - start
    m, s = divmod(time_taken, 60)
    h, m = divmod(m, 60)
    print('\nTime taken was {} : {} : {}.\n\n'.format(
                                                    int(h), int(m), s))


def vector_theta(vector1, vector2):
    """
    Returns the angle between two vectors in radians.

    Parameters
    ----------
    vector1 : numpy.array
        The first vector.

    vector2 : numpy.array
        The second vector.

    Returns
    -------
    float
        The angle between `vector1` and `vector2` in radians.

    """

    numerator = np.dot(vector1, vector2)
    denominator = (np.linalg.norm(vector1) *
                   np.linalg.norm(vector2))
    # This if statement prevents returns of NaN due to floating point
    # incurracy.
    if np.isclose(numerator, denominator, atol=1e-8):
        return 0.0
    return np.arccos(numerator/denominator)
